@@ -29,9 +29,11 @@ module Data.Parameterized.Map
   , insertWith
   , delete
   , union
+  , intersectWithKeyMaybe
     -- * Query
   , null
   , lookup
+  , findWithDefault
   , member
   , notMember
   , size
@@ -40,6 +42,8 @@ module Data.Parameterized.Map
   , elems
   , fromList
   , toList
+  , toAscList
+  , toDescList
   , fromKeys
   , fromKeysM
    -- * Filter
@@ -53,6 +57,8 @@ module Data.Parameterized.Map
   , foldrWithKey
   , foldrWithKey'
   , foldMapWithKey
+  , foldlMWithKey
+  , foldrMWithKey
     -- * Traversals
   , map
   , mapWithKey
@@ -60,11 +66,13 @@ module Data.Parameterized.Map
   , mapMaybeWithKey
   , traverseWithKey
   , traverseWithKey_
+  , traverseMaybeWithKey
     -- * Complex interface.
   , UpdateRequest(..)
   , Updated(..)
   , updatedValue
   , updateAtKey
+  , mergeWithKey
   , mergeWithKeyM
   , module Data.Parameterized.Classes
     -- * Pair
@@ -171,7 +179,6 @@ instance (TestEquality k, EqF a) => Eq (MapF k a) where
  #-}
 #endif
 
-
 -- | Apply function to all elements in map.
 mapWithKey
   :: (forall tp . ktp tp -> f tp -> g tp)
@@ -199,7 +206,8 @@ mapMaybe f = mapMaybeWithKey (\_ x -> f x)
 -- | Traverse elements in a map
 traverse :: Applicative m => (forall tp . f tp -> m (g tp)) -> MapF ktp f -> m (MapF ktp g)
 traverse _ Tip = pure Tip
-traverse f (Bin sx kx x l r) = Bin sx kx <$> f x <*> traverse f l <*> traverse f r
+traverse f (Bin sx kx x l r) =
+  (\l' x' r' -> Bin sx kx x' l' r') <$> traverse f l <*> f x <*> traverse f r
 
 -- | Traverse elements in a map
 traverseWithKey
@@ -209,7 +217,7 @@ traverseWithKey
   -> m (MapF ktp g)
 traverseWithKey _ Tip = pure Tip
 traverseWithKey f (Bin sx kx x l r) =
-   Bin sx kx <$> f kx x <*> traverseWithKey f l <*> traverseWithKey f r
+   (\l' x' r' -> Bin sx kx x' l' r') <$> traverseWithKey f l <*> f kx x <*> traverseWithKey f r
 
 -- | Traverse elements in a map without returning result.
 traverseWithKey_
@@ -217,20 +225,34 @@ traverseWithKey_
   => (forall tp . ktp tp -> f tp -> m ())
   -> MapF ktp f
   -> m ()
-traverseWithKey_ _ Tip = pure ()
-traverseWithKey_ f (Bin _ kx x l r) = f kx x *> traverseWithKey_ f l *> traverseWithKey_ f r
+traverseWithKey_ = \f -> foldrWithKey (\k v r -> f k v *> r) (pure ())
+{-# INLINABLE traverseWithKey_ #-}
 
+-- | Traverse keys\/values and collect the 'Just' results.
+traverseMaybeWithKey :: Applicative f
+                     => (forall tp . k tp -> a tp -> f (Maybe (b tp)))
+                     -> MapF k a -> f (MapF k b)
+traverseMaybeWithKey _ Tip = pure Tip
+traverseMaybeWithKey f (Bin _ kx x Tip Tip) = maybe Tip (\x' -> Bin 1 kx x' Tip Tip) <$> f kx x
+traverseMaybeWithKey f (Bin _ kx x l r) =
+    liftA3 combine (traverseMaybeWithKey f l) (f kx x) (traverseMaybeWithKey f r)
+  where
+    combine l' mx r' = seq l' $ seq r' $
+      case mx of
+        Just x' -> Bin.link (Pair kx x') l' r'
+        Nothing -> Bin.merge l' r'
+{-# INLINABLE traverseMaybeWithKey #-}
 
 type instance IndexF   (MapF k v) = k
 type instance IxValueF (MapF k v) = v
 
 -- | Turn a map key into a traversal that visits the indicated element in the map, if it exists.
-instance forall (k:: a -> Type) v. OrdF k => IxedF a (MapF k v) where
+instance forall a (k:: a -> Type) v. OrdF k => IxedF a (MapF k v) where
   ixF :: k x -> Traversal' (MapF k v) (v x)
   ixF i f m = updatedValue <$> updateAtKey i (pure Nothing) (\x -> Set <$> f x) m
 
 -- | Turn a map key into a lens that points into the indicated position in the map.
-instance forall (k:: a -> Type) v. OrdF k => AtF a (MapF k v) where
+instance forall a (k:: a -> Type) v. OrdF k => AtF a (MapF k v) where
   atF :: k x -> Lens' (MapF k v) (Maybe (v x))
   atF i f m = updatedValue <$> updateAtKey i (f Nothing) (\x -> maybe Delete Set <$> f (Just x)) m
 
@@ -248,6 +270,20 @@ lookup k0 = seq k0 (go k0)
         EQF -> Just x
 {-# INLINABLE lookup #-}
 
+-- | @findWithDefault d k m@ returns the value bound to @k@ in the map @m@, or @d@
+-- if @k@ is not bound in the map.
+findWithDefault :: OrdF k => a tp -> k tp -> MapF k a -> a tp
+findWithDefault = \def k -> seq k (go def k)
+  where
+    go :: OrdF k => a tp -> k tp -> MapF k a -> a tp
+    go d _ Tip = d
+    go d k (Bin _ kx x l r) =
+      case compareF k kx of
+        LTF -> go d k l
+        GTF -> go d k r
+        EQF -> x
+{-# INLINABLE findWithDefault #-}
+
 -- | Return true if key is bound in map.
 member :: OrdF k => k tp -> MapF k a -> Bool
 member k0 = seq k0 (go k0)
@@ -308,9 +344,7 @@ foldlWithKey' f z (Bin _ kx x l r) =
 foldrWithKey :: (forall s . k s -> a s -> b -> b) -> b -> MapF k a -> b
 foldrWithKey _ z Tip = z
 foldrWithKey f z (Bin _ kx x l r) =
-  let rz = foldrWithKey f z r
-      kz = f kx x rz
-   in foldrWithKey f kz l
+  foldrWithKey f (f kx x (foldrWithKey f z r)) l
 
 -- | Perform a strict right fold with the key also provided.
 foldrWithKey' :: (forall s . k s -> a s -> b -> b) -> b -> MapF k a -> b
@@ -325,6 +359,15 @@ foldMapWithKey :: Monoid m => (forall s . k s -> a s -> m) -> MapF k a -> m
 foldMapWithKey _ Tip = mempty
 foldMapWithKey f (Bin _ kx x l r) = foldMapWithKey f l <> f kx x <> foldMapWithKey f r
 
+-- | A monadic left-to-right fold over keys and values in the map.
+foldlMWithKey :: Monad m => (forall s . b -> k s -> a s -> m b) -> b -> MapF k a -> m b
+foldlMWithKey f z0 m = foldrWithKey (\k a r z ->  f z k a >>= r) pure m z0
+
+-- | A monadic right-to-left fold over keys and values in the map.
+foldrMWithKey :: Monad m => (forall s . k s -> a s -> b -> m b) -> b -> MapF k a -> m b
+foldrMWithKey f z0 m = foldlWithKey (\r k a z ->  f k a z >>= r) pure m z0
+
+-- | Pretty print keys and values in map.
 showMap :: (forall tp . ktp tp -> String)
         -> (forall tp . rtp tp -> String)
         -> MapF ktp rtp
@@ -406,7 +449,9 @@ delete = \k m -> seq k $ fromMaybeS m $ Bin.delete (p k) m
 {-# INLINABLE delete #-}
 {-# SPECIALIZE Bin.delete :: (Pair k a -> Ordering) -> MapF k a -> MaybeS (MapF k a) #-}
 
--- | Union two sets
+-- | Left-biased union of two maps. The resulting map will contain the
+-- union of the keys of the two arguments. When a key is contained in
+-- both maps the value from the first map will be preserved.
 union :: OrdF k => MapF k a -> MapF k a -> MapF k a
 union t1 t2 = Bin.union comparePairKeys t1 t2
 {-# INLINABLE union #-}
@@ -481,12 +526,21 @@ updateAtKey k onNotFound onFound t = ins <$> atKey' k onNotFound onFound t
 fromList :: OrdF k => [Pair k a] -> MapF k a
 fromList = foldl' (\m (Pair k a) -> insert k a m) Data.Parameterized.Map.empty
 
+-- | Return list of key-values pairs in map in ascending order.
+toAscList :: MapF k a -> [Pair k a]
+toAscList = foldrWithKey (\k x l -> Pair k x : l) []
+
+-- | Return list of key-values pairs in map in descending order.
+toDescList :: MapF k a -> [Pair k a]
+toDescList = foldlWithKey (\l k x -> Pair k x : l) []
+
+-- | Return list of key-values pairs in map.
 toList :: MapF k a -> [Pair k a]
-toList = foldrWithKey (\k x m -> Pair k x : m) []
+toList = toAscList
 
 -- | Generate a map from a foldable collection of keys and a
 -- function from keys to values.
-fromKeys :: forall m (t :: Type -> Type) (a :: k -> Type) (v :: k -> Type)
+fromKeys :: forall k m (t :: Type -> Type) (a :: k -> Type) (v :: k -> Type)
           .  (Monad m, Foldable t, OrdF a)
             => (forall tp . a tp -> m (v tp))
             -- ^ Function for evaluating a register value.
@@ -499,7 +553,7 @@ fromKeys f = foldM go empty
 
 -- | Generate a map from a foldable collection of keys and a monadic
 -- function from keys to values.
-fromKeysM :: forall m (t :: Type -> Type) (a :: k -> Type) (v :: k -> Type)
+fromKeysM :: forall k m (t :: Type -> Type) (a :: k -> Type) (v :: k -> Type)
           .  (Monad m, Foldable t, OrdF a)
            => (forall tp . a tp -> m (v tp))
            -- ^ Function for evaluating an input value to store the result in the map.
@@ -518,7 +572,72 @@ filterLtMaybe :: OrdF k => MaybeS (k x) -> MapF k a -> MapF k a
 filterLtMaybe NothingS m = m
 filterLtMaybe (JustS k) m = filterLt k m
 
--- | Merge bindings in two maps to get a third.
+-- | Returns only entries that are strictly between the two keys.
+filterMiddle :: OrdF k => k x -> k y -> MapF k a -> MapF k a
+filterMiddle lo hi (Bin _ k _ _ r)
+  | k `leqF` lo = filterMiddle lo hi r
+filterMiddle lo hi (Bin _ k _ l _)
+  | k `geqF` hi = filterMiddle lo hi l
+filterMiddle _  _  t = t
+{-# INLINABLE filterMiddle #-}
+
+{--------------------------------------------------------------------
+  [trim blo bhi t] trims away all subtrees that surely contain no
+  values between the range [blo] to [bhi]. The returned tree is either
+  empty or the key of the root is between @blo@ and @bhi@.
+--------------------------------------------------------------------}
+trim :: OrdF k => MaybeS (k x) -> MaybeS (k y) -> MapF k a -> MapF k a
+trim NothingS   NothingS   t = t
+trim (JustS lk) NothingS   t = filterGt lk t
+trim NothingS   (JustS hk) t = filterLt hk t
+trim (JustS lk) (JustS hk) t = filterMiddle lk hk t
+
+-- Helper function for 'mergeWithKeyM'. The @'trimLookupLo' lk hk t@ performs both
+-- @'trim' (JustS lk) hk t@ and @'lookup' lk t@.
+
+-- See Note: Type of local 'go' function
+trimLookupLo :: OrdF k => k tp -> MaybeS (k y) -> MapF k a -> Bin.PairS (Maybe (a tp)) (MapF k a)
+trimLookupLo lk NothingS t = greater lk t
+  where greater :: OrdF k => k tp -> MapF k a -> Bin.PairS (Maybe (a tp)) (MapF k a)
+        greater lo t'@(Bin _ kx x l r) =
+           case compareF lo kx of
+             LTF -> Bin.PairS (lookup lo l) t'
+             EQF -> Bin.PairS (Just x) r
+             GTF -> greater lo r
+        greater _ Tip = Bin.PairS Nothing Tip
+trimLookupLo lk (JustS hk) t = middle lk hk t
+  where middle :: OrdF k => k tp -> k y -> MapF k a -> Bin.PairS (Maybe (a tp)) (MapF k a)
+        middle lo hi t'@(Bin _ kx x l r) =
+          case compareF lo kx of
+            LTF | kx `ltF` hi -> Bin.PairS (lookup lo l) t'
+                | otherwise -> middle lo hi l
+            EQF -> Bin.PairS (Just x) (lesser hi r)
+            GTF -> middle lo hi r
+        middle _ _ Tip = Bin.PairS Nothing Tip
+
+        lesser :: OrdF k => k y -> MapF k a -> MapF k a
+        lesser hi (Bin _ k _ l _) | k `geqF` hi = lesser hi l
+        lesser _ t' = t'
+
+-- | Merge bindings in two maps using monadic actions to get a third.
+--
+-- The first function is used to merge elements that occur under the
+-- same key in both maps. Return Just to add an entry into the
+-- resulting map under this key or Nothing to remove this key from the
+-- resulting map.
+--
+-- The second function will be applied to submaps of the first map
+-- argument where no keys overlap with the second map argument. The
+-- result of this function must be a map with a subset of the keys of
+-- its argument.  This means the function can alter the values of its
+-- argument and it can remove key-value pairs from it, but it can
+-- break `MapF` ordering invariants if it introduces new keys.
+--
+-- Third function is analogous to the second function except that it applies
+-- to the second map argument of 'mergeWithKeyM' instead of the first.
+--
+-- Common examples of the two functions include 'id' when constructing a union
+-- or 'const' 'empty' when constructing an intersection.
 mergeWithKeyM :: forall k a b c m
                . (Applicative m, OrdF k)
               => (forall tp . k tp -> a tp -> b tp -> m (Maybe (c tp)))
@@ -555,53 +674,47 @@ mergeWithKeyM f g1 g2 = go
                           <*> hedgeMerge blo bmi l (trim blo bmi t2)
                           <*> hedgeMerge bmi bhi r trim_t2
       where bmi = JustS kx
+
 {-# INLINABLE mergeWithKeyM #-}
 
-{--------------------------------------------------------------------
-  [trim blo bhi t] trims away all subtrees that surely contain no
-  values between the range [blo] to [bhi]. The returned tree is either
-  empty or the key of the root is between @blo@ and @bhi@.
---------------------------------------------------------------------}
-trim :: OrdF k => MaybeS (k x) -> MaybeS (k y) -> MapF k a -> MapF k a
-trim NothingS   NothingS   t = t
-trim (JustS lk) NothingS   t = filterGt lk t
-trim NothingS   (JustS hk) t = filterLt hk t
-trim (JustS lk) (JustS hk) t = filterMiddle lk hk t
+-- | Merge bindings in two maps to get a third.
+--
+-- The first function is used to merge elements that occur under the
+-- same key in both maps. Return Just to add an entry into the
+-- resulting map under this key or Nothing to remove this key from the
+-- resulting map.
+--
+-- The second function will be applied to submaps of the first map
+-- argument where no keys overlap with the second map argument. The
+-- result of this function must be a map with a subset of the keys of
+-- its argument.  This means the function can alter the values of its
+-- argument and it can remove key-value pairs from it, but it can
+-- break `MapF` ordering invariants if it introduces new keys.
+--
+-- Third function is analogous to the second function except that it applies
+-- to the second map argument of 'mergeWithKeyM' instead of the first.
+--
+-- Common examples of the two functions include 'id' when constructing a union
+-- or 'const' 'empty' when constructing an intersection.
+mergeWithKey :: forall k a b c
+               . OrdF k
+              => (forall tp . k tp -> a tp -> b tp -> Maybe (c tp))
+              -> (MapF k a -> MapF k c)
+              -> (MapF k b -> MapF k c)
+              -> MapF k a
+              -> MapF k b
+              -> MapF k c
+mergeWithKey f g1 g2 x y = runIdentity $
+  mergeWithKeyM (\k a b -> pure $! f k a b) (pure . g1) (pure . g2) x y
 
--- | Returns only entries that are strictly between the two keys.
-filterMiddle :: OrdF k => k x -> k y -> MapF k a -> MapF k a
-filterMiddle lo hi (Bin _ k _ _ r)
-  | k `leqF` lo = filterMiddle lo hi r
-filterMiddle lo hi (Bin _ k _ l _)
-  | k `geqF` hi = filterMiddle lo hi l
-filterMiddle _  _  t = t
-{-# INLINABLE filterMiddle #-}
-
-
-
--- Helper function for 'mergeWithKey'. The @'trimLookupLo' lk hk t@ performs both
--- @'trim' (JustS lk) hk t@ and @'lookup' lk t@.
-
--- See Note: Type of local 'go' function
-trimLookupLo :: OrdF k => k tp -> MaybeS (k y) -> MapF k a -> Bin.PairS (Maybe (a tp)) (MapF k a)
-trimLookupLo lk NothingS t = greater lk t
-  where greater :: OrdF k => k tp -> MapF k a -> Bin.PairS (Maybe (a tp)) (MapF k a)
-        greater lo t'@(Bin _ kx x l r) =
-           case compareF lo kx of
-             LTF -> Bin.PairS (lookup lo l) t'
-             EQF -> Bin.PairS (Just x) r
-             GTF -> greater lo r
-        greater _ Tip = Bin.PairS Nothing Tip
-trimLookupLo lk (JustS hk) t = middle lk hk t
-  where middle :: OrdF k => k tp -> k y -> MapF k a -> Bin.PairS (Maybe (a tp)) (MapF k a)
-        middle lo hi t'@(Bin _ kx x l r) =
-          case compareF lo kx of
-            LTF | kx `ltF` hi -> Bin.PairS (lookup lo l) t'
-                | otherwise -> middle lo hi l
-            EQF -> Bin.PairS (Just x) (lesser hi r)
-            GTF -> middle lo hi r
-        middle _ _ Tip = Bin.PairS Nothing Tip
-
-        lesser :: OrdF k => k y -> MapF k a -> MapF k a
-        lesser hi (Bin _ k _ l _) | k `geqF` hi = lesser hi l
-        lesser _ t' = t'
+-- | Applies a function to the pairwise common elements of two maps.
+--
+-- Formally, we have that @intersectWithKeyMaybe f x y@ contains a
+-- binding from a key @k@ to a value @v@ if and only if @x@ and @y@
+-- bind @k@ to @x_k@ and @y_k@ and @f x_k y_k = Just v@.
+intersectWithKeyMaybe :: OrdF k
+                      => (forall tp . k tp -> a tp -> b tp -> Maybe (c tp))
+                      -> MapF k a
+                      -> MapF k b
+                      -> MapF k c
+intersectWithKeyMaybe f = mergeWithKey f (const empty) (const empty)
